@@ -116,6 +116,7 @@ ALLOWED_NODE_TYPES = (
     _ast.Add,
     _ast.Sub,
     _ast.Num if hasattr(_ast, 'Num') else _ast.Constant,
+    _ast.List,
 
     _ast.BitAnd,
     _ast.BitOr,
@@ -475,11 +476,18 @@ def _evaluate_expression(expr, fields_and_datasets, variables, cache):
             expr.value, fields_and_datasets, variables, cache
         )
         # assume that we will only use a single index
-        slice_ = expr.slice.value.value
+        if isinstance(expr.slice.value, int):
+            slice_ = expr.slice.value  # python 3.10
+        else:
+            slice_ = expr.slice.value.value  # python 3.8
 
         # Get a copy of the selected band
         data = value.GetRasterBand(slice_ + 1).ReadAsArray()
         result = gdal_array.OpenNumPyArray(data, True)
+        # restore nodata on output
+        nodata_value = value.GetRasterBand(slice_ + 1).GetNoDataValue()
+        if nodata_value is not None:
+            result.GetRasterBand(1).SetNoDataValue(nodata_value)
 
     elif hasattr(_ast, 'Num') and isinstance(expr, _ast.Num):
         result = expr.n
@@ -487,6 +495,12 @@ def _evaluate_expression(expr, fields_and_datasets, variables, cache):
     elif hasattr(_ast, 'Constant') and isinstance(expr, _ast.Constant):
         result = expr.value
 
+    elif hasattr(_ast, 'List') and isinstance(expr, _ast.List):
+        result = [
+            _evaluate_expression(
+                item, fields_and_datasets, variables, cache,
+            ) for item in expr.elts
+        ]
     else:
         raise BandExpressionError('Invalid expression node %s' % expr)
 
