@@ -26,6 +26,7 @@
 # ------------------------------------------------------------------------------
 
 from os.path import join
+from typing import List, Type, Iterable, Tuple
 from uuid import uuid4
 try:
     from itertools import izip_longest
@@ -47,8 +48,10 @@ from eoxserver.render.browse.generate import (
 from eoxserver.render.map.objects import (
     CoverageLayer, CoveragesLayer, MosaicLayer, OutlinedCoveragesLayer,
     BrowseLayer, OutlinedBrowseLayer,
-    MaskLayer, MaskedBrowseLayer, OutlinesLayer
+    MaskLayer, MaskedBrowseLayer, OutlinesLayer,
+    Layer, Map,
 )
+from eoxserver.render.coverage.objects import Coverage, Field
 from eoxserver.render.mapserver.config import (
     DEFAULT_EOXS_MAPSERVER_LAYER_FACTORIES,
 )
@@ -64,23 +67,25 @@ logger = logging.getLogger(__name__)
 
 
 class BaseMapServerLayerFactory(object):
-    handled_layer_types = []
+    handled_layer_types: List[Type[Layer]] = []
 
     @classmethod
-    def supports(self, layer_type):
+    def supports(self, layer_type: Type[Layer]):
         return layer_type in self.handled_layer_types
 
-    def create(self, map_obj, layer):
+    def create(self, map_obj: Map, layer: Layer):
         pass
 
-    def destroy(self, map_obj, layer, data):
+    def destroy(self, map_obj: Map, layer: Layer, data):
         pass
 
 
 class CoverageLayerFactoryMixIn(object):
     """ Base class for factories dealing with coverages.
     """
-    def get_fields(self, fields, bands, wavelengths):
+    def get_fields(
+        self, fields: Iterable[Field], bands, wavelengths
+    ) -> List[Field]:
         """ Get the field subset for the given bands/wavelengths selection
         """
         if bands:
@@ -112,7 +117,7 @@ class CoverageLayerFactoryMixIn(object):
 
         return fields
 
-    def create_coverage_layer(self, map_obj, coverage, fields,
+    def create_coverage_layer(self, map_obj: Map, coverage: Coverage, fields: List[Field],
                               style=None, ranges=None):
         """ Creates a mapserver layer object for the given coverage
         """
@@ -132,8 +137,8 @@ class CoverageLayerFactoryMixIn(object):
         # TODO: apply subsets in time/elevation dims
         num_locations = len(set(locations))
         if num_locations == 1:
+            location = field_locations[0][1]
             if not coverage.grid.is_referenceable:
-                location = field_locations[0][1]
                 data = location.path
                 ms.set_env(map_obj, location.env, True)
             else:
@@ -145,12 +150,13 @@ class CoverageLayerFactoryMixIn(object):
                 wkt = osr.SpatialReference(map_obj.getProjection()).wkt
 
                 # TODO: env?
-                reftools.create_rectified_vrt(
-                    field_locations[0][1].path, vrt_path,
-                    order=1, max_error=10,
-                    resolution=(resx, -resy), srid_or_wkt=wkt
-                )
-                data = vrt_path
+                with gdal.config_env(location.env):
+                    reftools.create_rectified_vrt(
+                        location.path, vrt_path,
+                        order=1, max_error=10,
+                        resolution=(resx, -resy), srid_or_wkt=wkt
+                    )
+                    data = vrt_path
 
         elif num_locations > 1:
             paths_set = set(
@@ -270,7 +276,7 @@ class OutlinedCoverageLayerFactory(CoverageLayerFactoryMixIn,
                                    BaseMapServerLayerFactory):
     handled_layer_types = [OutlinedCoveragesLayer]
 
-    def create(self, map_obj, layer):
+    def create(self, map_obj, layer: CoveragesLayer):
         coverages = layer.coverages
         style = layer.style
 
@@ -311,7 +317,7 @@ class OutlinedCoverageLayerFactory(CoverageLayerFactoryMixIn,
 class MosaicLayerFactory(CoverageLayerFactoryMixIn, BaseMapServerLayerFactory):
     handled_layer_types = [MosaicLayer]
 
-    def create(self, map_obj, layer):
+    def create(self, map_obj, layer: MosaicLayer):
         mosaic = layer.mosaic
         fields = self.get_fields(
             mosaic.range_type, layer.bands, layer.wavelengths
@@ -344,7 +350,8 @@ class BrowseLayerMixIn(object):
                         map_.width, map_.height,
                         map_.bbox,
                         map_.crs,
-                        filename_generator
+                        filename_generator,
+                        browse.variables,
                     )
                 layer_objs = _create_raster_layer_objs(
                     map_obj, browse.extent, browse.spatial_reference,
@@ -394,9 +401,9 @@ class BrowseLayerMixIn(object):
 
                 else:
                     browse_iter = enumerate(
-                        zip(browse.field_list, browse.ranges), start=1
+                        zip(browse.field_list, browse.ranges, browse.nodata_values), start=1
                     )
-                    for i, (field, field_range) in browse_iter:
+                    for i, (field, field_range, nodata_value) in browse_iter:
                         if ranges:
                             if len(ranges) == 1:
                                 range_ = ranges[0]
@@ -408,10 +415,34 @@ class BrowseLayerMixIn(object):
                             range_ = _get_range(field)
 
                         for layer_obj in layer_objs:
-                            layer_obj.setProcessingKey(
-                                "SCALE_%d" % i,
-                                "%s,%s" % tuple(range_)
-                            )
+                            # NOTE: Only works if browsetype nodata is lower than browse_type_min by at least 1
+                            if browse.show_out_of_bounds_data:
+                                # final LUT for min,max 200,700 and nodata=0 should look like:
+                                # 0:0,1:1,200:1,700:256
+                                lut_inputs = {
+                                    range_[0]: 1,
+                                    range_[1]: 256,
+                                }
+                                if nodata_value is not None:
+                                    # no_data_value_plus +1 to ensure that only no_data_value is
+                                    # rendered as black (transparent)
+                                    nodata_value_plus = nodata_value + 1
+                                    lut_inputs[nodata_value] = 0
+                                    lut_inputs[nodata_value_plus] = 1
+
+                                # LUT inputs needs to be ascending
+                                sorted_inputs = {
+                                    k: v for k, v in sorted(list(lut_inputs.items()))
+                                }
+                                lut = ",".join("%d:%d" % (k,v) for k,v in sorted_inputs.items())
+
+                                layer_obj.setProcessingKey("LUT_%d" % i, lut)
+                            else:
+                                # due to offsite 0,0,0 will make all pixels below or equal to min transparent
+                                layer_obj.setProcessingKey(
+                                    "SCALE_%d" % i,
+                                    "%s,%s" % tuple(range_)
+                                )
 
             elif isinstance(browse, Browse):
                 layer_objs = _create_raster_layer_objs(
@@ -631,7 +662,7 @@ class OutlinesLayerFactory(BaseMapServerLayerFactory):
 
 
 def _create_raster_layer_objs(map_obj, extent, sr, data, filename_generator,
-                              resample=None):
+                              resample=None) -> List[ms.layerObj]:
     layer_obj = ms.layerObj(map_obj)
     layer_obj.type = ms.MS_LAYER_RASTER
     layer_obj.status = ms.MS_ON
@@ -852,7 +883,7 @@ def _create_raster_style(name, layer, minvalue=0, maxvalue=255,
         layer.insertClass(cls)
 
 
-def _get_range(field, range_=None):
+def _get_range(field: Field, range_=None) -> Tuple[int, int]:
     """ Gets the numeric range of a field
     """
     if range_:
@@ -884,7 +915,7 @@ def _setup_factories():
     ]
 
 
-def get_layer_factories():
+def get_layer_factories() -> List[BaseMapServerLayerFactory]:
     if LAYER_FACTORIES is None:
         _setup_factories()
     return LAYER_FACTORIES
