@@ -42,7 +42,6 @@ from django.contrib.gis.geos import Polygon
 from django.db.models import Min, Max, Q, F, ExpressionWrapper
 from django.db.models.functions import Cast
 from django.utils.timezone import now
-from django.utils.encoding import python_2_unicode_compatible
 from model_utils.managers import InheritanceManager
 from jsonfield import JSONField
 
@@ -202,6 +201,8 @@ class BrowseType(models.Model):
     blue_range_max = models.FloatField(**optional)
     alpha_range_max = models.FloatField(**optional)
 
+    show_out_of_bounds_data = models.BooleanField(default=False, **mandatory)
+
     def __str__(self):
         if self.name:
             return self.name
@@ -314,7 +315,6 @@ def eo_object_identifier_validator(value):
         identifier_validators[0](value)
 
 
-@python_2_unicode_compatible
 class EOObject(models.Model):
     """ Base class for Collections, Products and Coverages
     """
@@ -530,7 +530,6 @@ class CollectionMetadata(models.Model):
 # ==============================================================================
 
 
-@python_2_unicode_compatible
 class AbstractCommonValue(models.Model):
     value = models.CharField(max_length=256, db_index=True, unique=True)
 
